/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/insertprogress.c
 * @brief handles file insertions in the GTK GUI
 * @author Igor Wronsky
 * @author Christian Grothoff (refactoring, added bugs)
 **/

#include "config.h"

#include <errno.h>
#include <gtk/gtk.h>
#include <string.h>
#include <unistd.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/gtkui/insertprogress.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/encoding/keywords.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/insert.h"
#include "applications/afs/gtkui/main.h"
#include "util/port.h"
#include "util/storage.h"
#include "util/logging.h"


/**
 * Callback function to update the insert progressbar.
 *
 * @param stats Statistics related to insert progression
 * @param ilm Data related to the insertion
 **/
static void insertModelCallback(ProgressStats * stats,
				InsertModel * ilm) {
  gdk_threads_enter();
  gtk_progress_set_value(GTK_PROGRESS(ilm->progressBar),
			 stats->progress);
  gdk_threads_leave();
}

/**
 * A function to be run by the insert thread. Does the
 * actual insertion.
 *
 * @param ilm Collected data related to the insertion
 */
void insertFileGtkThread(InsertModel * ilm) {
  int res;
  GNUNET_TCP_SOCKET * sock;
  Block * top;
  int i;
  
  SEMAPHORE_DOWN(refuseToDie);
  sock = getClientSocket();
  top = insertFile(sock,
		   ilm->fileName,
		   (ProgressModel)&insertModelCallback,
		   ilm);
  if (top != NULL) {
    res = insertRoot(sock,
		     top, 
		     ilm->description,
		     ilm->fileNameRoot,
		     ilm->mimetype,
		     ilm->num_keywords,
		     ilm->keywords,
		     NULL);
  } else {
    res = SYSERR;
  }
  gdk_threads_enter();
  gtk_widget_destroy(ilm->progressBarWindow);
  if (res == OK) {
    HexName hex;
    HexName hex2;
    
    hash2hex(&top->chk.key,&hex);
    hash2hex(&top->chk.query, &hex2);

    guiMessage("Successfully processed %s\n%s\n%s\n%u %u\n", 
	       ilm->fileName,
	       (char*)&hex,
	       (char*)&hex2,
	       crc32N(top->data, top->len),
	       top->filesize);
    LOG(LOG_DEBUG,
	"DEBUG: Successfully processed %s -- %s %s %u %u\n", 
	ilm->fileName,
	(char*)&hex,
	(char*)&hex2,
	crc32N(top->data, top->len),
	top->filesize);
  } else {
    guiMessage("Insertion of %s FAILED!\n", 
	       ilm->fileName);
  }
  gdk_threads_leave();
  if(top != NULL)
    top->done(top, NULL);
  releaseClientSocket(sock);

  /* insert complete */
  SEMAPHORE_UP(refuseToDie);
  for (i=0;i<ilm->num_keywords;i++)
    FREE(ilm->keywords[i]);
  FREENONNULL(ilm->keywords);
  if(ilm->deleteAfterInsert == YES) 
    unlink(ilm->fileName);
  FREE(ilm->fileName);
  FREE(ilm->mimetype);
  FREE(ilm->description);
  FREE(ilm->fileNameRoot);
  FREE(ilm);
}

/**
 * Callback for handling "delete_event": keep the window OPEN.
 **/
gint refuseDeleteEvent(GtkWidget * widget,
		       GdkEvent * event,
		       gpointer data) {
  LOG(LOG_DEBUG, 
      "DEBUG: refuseDeleteEvent\n");
  return TRUE;
}


void createInsertProgressBar(InsertModel * ilm) {
  GtkWidget * window;
  GtkWidget * box;
  GtkObject * adjustment;
  int fileLength;

  /* create a new window for a progressbar */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  ilm->progressBarWindow = window;
  gtk_window_set_title(GTK_WINDOW(window), 
		       ilm->fileName);
  box = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), 
		    box);
  gtk_signal_connect(GTK_OBJECT(window),
		     "delete_event",
		     GTK_SIGNAL_FUNC(refuseDeleteEvent),
		     NULL);
  gtk_container_set_border_width(GTK_CONTAINER(window), 
				 10);

 
  /* create the actual progressbar */
  fileLength = getFileSize(ilm->fileName);
  ilm->progressBar = gtk_progress_bar_new();
  gtk_progress_set_show_text(GTK_PROGRESS(ilm->progressBar),
			     1);
  gtk_progress_set_format_string(GTK_PROGRESS(ilm->progressBar),
				 (gchar*)"%v bytes inserted");
  adjustment = gtk_adjustment_new(0,
				  0,
				  fileLength,
				  1,
				  0,
				  0);
  gtk_progress_set_adjustment(GTK_PROGRESS(ilm->progressBar),
			      GTK_ADJUSTMENT(adjustment));
  gtk_box_pack_start(GTK_BOX(box),
		     ilm->progressBar, 
		     TRUE, 
		     TRUE, 
		     0);
  gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(ilm->progressBar),
				   GTK_PROGRESS_LEFT_TO_RIGHT);
  gtk_widget_show(ilm->progressBar);
  gtk_widget_show(box);
  gtk_widget_show(window);
}

/* end of insertprogress.c */
