/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file server/heloexchange.c
 * @brief Cron-jobs that exchange HELOs to ensure that the network is
 * connected (nodes know of each other).
 *
 * @author Christian Grothoff
 **/

#include "config.h"

#ifndef CYGWIN
 #include <sys/socket.h>
#else
 #include <cygwin/socket.h>
 #include <fcntl.h>
#endif
#ifdef SOLARIS
 #include <signal.h>
#endif
#include <errno.h>
#include <math.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/statistics.h"
#include "util/xmalloc.h"
#include "util/storage.h"
#include "server/heloexchange.h"
#include "server/keyservice.h"
#include "server/knownhosts.h"
#include "server/pingpong.h"
#include "server/handler.h"

#define TCP_HTTP_PORT 80
#define HTTP_URL "http://"
#define GET_COMMAND "GET http://%s%s HTTP/1.0\r\n\r\n"

#define HELO_BROADCAST_FREQUENCY 5 * cronMINUTES
#define HELO_FORWARD_FREQUENCY 7 * cronMINUTES

#define HELO_HELPER_TABLE_START_SIZE 128

#define DEBUG_HELOEXCHANGE NO

/* ************* internal Methods **************** */

/**
 * Tell everybody we are there...
 **/
static void broadcastHELO(void * unused);
 
/**
 * Forward HELOs from all known hosts to all known hosts.
 **/
static void forwardHELO(void * unused);

/* handles for stats */
static int stat_helo_received_via_http;
static int stat_helo_received;
static int stat_helo_valid_received;
static int stat_helo_forwarded;
static int stat_helo_initiated;

/* how many helos did we receive so far via http? 
   used to spread out the cron-jobs... */
static int httpHeloCount = 0;

/**
 * The HTTP proxy (optional)
 **/
static struct sockaddr_in theProxy;

/* ******************** CODE ********************* */


/**
 * Catch sigpipe on send, and do nothing
 **/
#ifndef LINUX
static void catcher(int sig) {
  LOG(LOG_INFO, 
      "INFO: signal %d caught\n", sig);
  signal(sig, catcher);
}
#endif

/**
 * Type for a HELO send via an encrypted channel.
 **/
int eHELOHandler(HostIdentity * sender,
		 p2p_HEADER * message) {
  receivedHELO(message);
  return OK; /* even if we had errors processing the HELO, keep going */
}


/**
 * initialize a few cron jobs. Must be called after
 * initcron (!).
 **/
void initHeloExchange() {
  char *proxy, *proxyPort;
  struct hostent *ip;

  stat_helo_received_via_http 
    = statHandle("# HELO messages received from http server");
  stat_helo_received
    = statHandle("# HELO messages received overall");
  stat_helo_valid_received
    = statHandle("# valid HELO messages received");
  stat_helo_forwarded
    = statHandle("# HELO messages forwarded from other peers");
  stat_helo_initiated
    = statHandle("# HELO messages originated");

  registerp2pHandler(p2p_PROTO_HELO,
		     &eHELOHandler);
  addCronJob(&broadcastHELO,
	     1 * cronMINUTES, 
	     HELO_BROADCAST_FREQUENCY,
	     NULL); 
  if (testConfigurationString("NETWORK",
			      "HELOEXCHANGE",
			      "YES") == YES)
    addCronJob(&forwardHELO,
	       2 * cronMINUTES, 
	       HELO_FORWARD_FREQUENCY,
	       NULL); 
#if DEBUG_HELOEXCHANGE
  else
    LOG(LOG_DEBUG, 
	"DEBUG: HELO forwarding disabled!\n");
#endif
  proxy = getConfigurationString("GNUNETD", "HTTP-PROXY");
  if (proxy != NULL) {
    ip = gethostbyname(proxy);
    if (ip == NULL) {
      LOG(LOG_ERROR, 
	  "Couldn't resolve name of HTTP proxy %s\n",
	  proxy);
      theProxy.sin_addr.s_addr = 0;
    } else {
      theProxy.sin_addr.s_addr = ((struct in_addr *)ip->h_addr)->s_addr;
      proxyPort = getConfigurationString("GNUNETD",
					 "HTTP-PROXY-PORT");
      if (proxyPort == NULL) {
	theProxy.sin_port = htons(8080);
      } else {
	theProxy.sin_port = htons(atoi(proxyPort));
	FREE(proxyPort);
      }
    }
    FREE(proxy);
  } else {
    theProxy.sin_addr.s_addr = 0;
  }
}

/**
 * Stops a few cron jobs that exchange HELOs.
 **/
void doneHeloExchange() {
  delCronJob(&broadcastHELO,
	     HELO_BROADCAST_FREQUENCY,
	     NULL); 
  if (testConfigurationString("NETWORK",
			      "HELOEXCHANGE",
			      "YES") == YES)
    delCronJob(&forwardHELO,
	       HELO_FORWARD_FREQUENCY,
	       NULL); /* seven minutes: exchange */
}

/**
 * We have received a HELO. Verify (signature, integrity,
 * ping-pong) and store identity if ok.
 *
 * @param message the HELO message
 * @return SYSERR on error, OK on success
 **/
int receivedHELO(p2p_HEADER * message) {
  TSession * tsession;
  HELO_Message * copy;
  HostIdentity foreignId;
  HELO_Message * msg;
  char * buffer;    
  int heloEnd;
  int mtu;
  int res;

  /* first verify that it is actually a valid HELO */
  msg = (HELO_Message* ) message;
  statChange(stat_helo_received, 1);
  if (ntohs(msg->header.size) != HELO_Message_size(msg))
    return SYSERR;
  getHostIdentity(&msg->publicKey,
		  &foreignId);
  if (!equalsHashCode160(&msg->senderIdentity.hashPubKey,
			 &foreignId.hashPubKey))
    return SYSERR; /* public key and host hash do not match */
  if (SYSERR == verifySig(&msg->senderIdentity,
			  HELO_Message_size(msg) - sizeof(Signature) - sizeof(PublicKey) - sizeof(p2p_HEADER),
			  &msg->signature,
			  &msg->publicKey)) {
    LOG(LOG_WARNING, 
	"WARNING: HELO message received invalid (signature bogous). Dropping.\n");
    return SYSERR; /* message invalid */  
  }
  if ((TIME_T)ntohl(msg->expirationTime) > TIME(NULL) + MAX_HELO_EXPIRES) {
     LOG(LOG_WARNING, 
	"WARNING: HELO message received invalid (expiration time over limit). Dropping.\n");   
    return SYSERR;
  }
  if (SYSERR == transportVerifyHelo(msg)) 
    return OK; /* not good, but do process rest of message */ 
  
  statChange(stat_helo_valid_received, 1);
#if DEBUG_HELOEXCHANGE
  LOG(LOG_INFO,
      "INFO: HELO advertisement for protocol %d received\n",
      ntohs(msg->protocol));
#endif
  /* Then check if we have seen this HELO before, if it is identical
     except for the TTL, we trust it and do not play PING-PONG */
  if (OK == identity2Helo(&foreignId,
			  ntohs(msg->protocol),
			  NO,
			  &copy) ) {
    if (0 == memcmp(&msg->senderAddressSize,
		    &copy->senderAddressSize,
		    sizeof(unsigned short)*2+
		    sizeof(unsigned int) + 
		    ntohs(copy->senderAddressSize)) ) {
      /* ok, we've seen this one exactly like this before (at most the
	 TTL has changed); thus we can 'trust' it without playing
	 ping-pong */
      bindAddress(msg);
      FREE(copy);
      return OK;
    } else {
#if DEBUG_HELOEXCHANGE
      LOG(LOG_DEBUG,
	  "DEBUG: advertised HELO differs from prior knowledge, requireing ping-pong confirmation\n");
      LOG(LOG_EVERYTHING,
	  "DEBUG: HELO-diff: %d -- %d, %d -- %d, %d -- %d, %d -- %d\n",
	  msg->senderAddressSize,
	  copy->senderAddressSize,
	  msg->protocol,
	  copy->protocol,
	  msg->MTU,
	  copy->MTU,
	  *(int*)&msg->senderAddress,
	  *(int*)&copy->senderAddress);
#endif
    }
    FREE(copy);
  }

  /* Ok, must play PING-PONG. Add the HELO to the temporary
     (in-memory only) buffer to make it available for a short
     time in order to play PING-PONG */
  copy = MALLOC(HELO_Message_size(msg));
  memcpy(copy,
	 msg,
	 HELO_Message_size(msg));
  addTemporaryHost(copy);


  /* Establish session as advertised in the HELO */
  copy = MALLOC(HELO_Message_size(msg));
  memcpy(copy,
	 msg,
	 HELO_Message_size(msg));
  if (SYSERR == transportConnect(copy, /* copy is freed by callee,
					  except on SYSERR! */
				 &tsession)) {
    FREE(copy);
    return SYSERR; /* could not connect */
  }    

  /* build message to send, ping must contain return-information,
     such as a selection of our HELOs... */
  mtu = transportGetMTU(tsession->ttype);
  buffer = MALLOC(mtu);
  heloEnd = getAdvertisedHELOs(mtu - sizeof(PINGPONG_Message),
			       buffer);
  if (heloEnd == -1) {
    LOG(LOG_WARNING,
	"WARNING: could not getAdvertisedHELOs, no PINGing\n");
    FREE(buffer);
    transportDisconnect(tsession);    
    return SYSERR;
  }
  copy = MALLOC(HELO_Message_size(msg));
  memcpy(copy,
	 msg,
	 HELO_Message_size(msg));
  res = OK;
  if (SYSERR == pingAction(&msg->senderIdentity,
			   (void (*)(void *))&bindAddress,
			   copy,
			   (PINGPONG_Message*)&buffer[heloEnd])) {
    FREE(copy);
    res = SYSERR;
    LOG(LOG_WARNING,
	"WARNING: could not send HELOs+PING, ping buffer full\n");
  }
  /* ok, finally we can send! */
  if (res == OK)
    if (SYSERR == transportSend(tsession,
				buffer,
				heloEnd + sizeof(PINGPONG_Message),
				NO, /* not encrypted */
				crc32N(buffer, heloEnd + sizeof(PINGPONG_Message))))
      res = SYSERR;
  FREE(buffer);
  if (SYSERR == transportDisconnect(tsession))
    res = SYSERR;
  return res;
}

static void receiveHeloDeferred(p2p_HEADER * msg) {
  receivedHELO(msg);
  FREE(msg);
}

/**
 * Download hostlist from the web.
 **/
static void downloadHostlistHelper(char * url) {
  unsigned short port;
  char * hostname;
  char * filename;
  unsigned int curpos;
  struct hostent *ip_info;
  struct sockaddr_in soaddr;
  int sock;
  int ret;
  char * command;
  cron_t start;
  char c;
  char * buffer;
  p2p_HEADER ** helos;
  int helosCount;
  int helosLen;
  int * perm;
  int i;

  port = TCP_HTTP_PORT;

#if DEBUG_HELOEXCHANGE
  LOG(LOG_INFO,
      "INFO: Trying to download a hostlist from %s\n",
      url);
#endif
  
  if (0 != strncmp(HTTP_URL, url, strlen(HTTP_URL)) ) {
    LOG(LOG_WARNING, 
	"WARNING: invalid URL %s (must begin with %s)\n", 
	url, 
	HTTP_URL);
    return;
  }
  curpos = strlen(HTTP_URL);
  hostname = &url[curpos];
  while ( (curpos < strlen(url)) &&
	  (url[curpos] != '/') )
    curpos++;
  if (curpos == strlen(url))
    filename = STRDUP("/");
  else 
    filename = STRDUP(&url[curpos]);
  url[curpos] = '\0'; /* terminator for hostname */  

  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    LOG(LOG_ERROR,
	"ERROR: could not open socket for hostlist download (%s).\n",
	strerror(errno));
    FREE(filename);
    return;
  }

  /* Do we need to connect through a proxy? */
  if (theProxy.sin_addr.s_addr == 0)
  {
   ip_info = gethostbyname(hostname);
   if (ip_info == NULL) {
    LOG(LOG_WARNING,
        "WARNING: could not download hostlist, host %s unknown\n",
        hostname);
    FREE(filename);
    return;
   }

   soaddr.sin_addr.s_addr = ((struct in_addr*)(ip_info->h_addr))->s_addr;
   soaddr.sin_port = htons(TCP_HTTP_PORT);
  }
  else
  {
   soaddr.sin_addr.s_addr = theProxy.sin_addr.s_addr;
   soaddr.sin_port = theProxy.sin_port;
  }
  soaddr.sin_family = AF_INET;

  if (connect (sock, 
	       (struct sockaddr*)&soaddr, 
	       sizeof(soaddr)) < 0) {
    LOG(LOG_WARNING,
	"WARNING: failed to send HTTP request to host %s (%u - %d) - (%s)\n",
	hostname,
	curpos, 
	sock,
	strerror(errno));
    FREE(filename);
    CLOSE(sock);
    return;
  }
  
  command = MALLOC(strlen(filename) + strlen(GET_COMMAND) + strlen(hostname));
  sprintf(command, GET_COMMAND, hostname, filename);
  FREE(filename);
  curpos = strlen(command)+1;
#ifndef LINUX
  if ( SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
    if (SIG_ERR == signal(SIGPIPE, catcher))
      LOG(LOG_WARNING, 
	  "WARNING: could not install handler for SIGPIPE!\n"\
	  "Attempting to continue anyway.");
  }
  curpos = send(sock, 
		command, 
		curpos, 
		0);
#else
  curpos = send(sock, 
		command,
		curpos, 
		MSG_NOSIGNAL);
#endif
  if (curpos != strlen(command)+1) {
    LOG(LOG_WARNING,
	"WARNING: failed so send HTTP request %s to host %s (%u - %d) - %s\n",
	command,
	hostname,
	curpos, 
	sock,
	strerror(errno));
    FREE(command);
    CLOSE(sock);
    return;
  }
  FREE(command);
  cronTime(&start);

  /* we first have to read out the http_response*/
  /* it ends with four line delimiters: "\r\n\r\n" */
  curpos = 0;
  while (curpos < 4) {
    if (start + 30 * cronSECONDS < cronTime(NULL))
      break; /* exit after 30s */
#ifndef CYGWIN
    ret = recv(sock,
	       &c, 
	       sizeof(c),
	       MSG_DONTWAIT);
#else
    /* enable non-blocking mode*/
    fcntl(sock, 
	  F_SETFL, 
	  fcntl(sock, F_GETFL) | O_NONBLOCK);
    ret = recv(sock,
	       &c, 
	       sizeof(c),
               0);
    fcntl(sock, 
	  F_SETFL, 
	  fcntl(sock, F_GETFL) & ~O_NONBLOCK);
#endif
    if (ret == 0)
      break;
    if (ret < 0) {
      if (errno != EAGAIN)
 	break;
      else
 	ret = 0;
    }
    if ((c=='\r') || (c=='\n')) 
      curpos += ret;
    else 
      curpos=0;    
}

if (curpos < 4) { /* we have not found it */
  LOG(LOG_WARNING, 
	"WARNING: exit downloadHostlist (error: no http response read)\n");
    CLOSE(sock);
    return;
  }

  helosCount = 0;
  helosLen = 0;
  helos = NULL;
  buffer = MALLOC(MAX_BUFFER_SIZE);
  while (1) {
    HELO_Message * helo;
    
    helo = (HELO_Message*) &buffer[0];
    helo->header.requestType = htons(p2p_PROTO_HELO);

    if (start + 30 * cronSECONDS < cronTime(NULL))
      break; /* exit after 30s */
    curpos = 0;
    helo->senderAddressSize = 0;
    while (curpos < HELO_Message_size(helo)) {
      if (start + 30 * cronSECONDS < cronTime(NULL))
	break; /* exit after 30s */
#ifndef CYGWIN
    ret = recv(sock, 
	       &((char*)helo)[curpos],
	       HELO_Message_size(helo)-curpos,
	       MSG_DONTWAIT);
#else
    /* enable non-blocking mode */
    fcntl(sock, 
	  F_SETFL,
	  fcntl(sock, F_GETFL) | O_NONBLOCK);
    ret = recv(sock, 
	       &((char*)helo)[curpos],
	       HELO_Message_size(helo)-curpos,
	       0);
    fcntl(sock,
	  F_SETFL, 
	  fcntl(sock, F_GETFL) & ~O_NONBLOCK);
#endif
      if (ret == 0)
	break;
      if (ret < 0) {
	if (errno != EAGAIN)
	  break;
	else
	  ret = 0;
      }
      if (HELO_Message_size(helo) >= MAX_BUFFER_SIZE) 
	break; /* INVALID! Avoid overflow! */      
      curpos += ret;
    }
    if (curpos != HELO_Message_size(helo)) {
      if (curpos != 0)
	LOG(LOG_WARNING,
	    "WARNING: received incomplete/invalid helo from http\n");
      break;
    }
    helo->header.size = htons(HELO_Message_size(helo));
#if DEBUG_HELOEXCHANGE
    LOG(LOG_CRON,
	".");
#endif
    statChange(stat_helo_received_via_http, 1);

    if(helosCount >= helosLen) 
      GROW(helos,
	   helosLen,
	   helosLen + HELO_HELPER_TABLE_START_SIZE);
    helos[helosCount++] = MALLOC(ntohs(helo->header.size));
    memcpy(helos[helosCount-1],
	   helo,
	   ntohs(helo->header.size));
    httpHeloCount++;  
  }
      
  /* process fetched helos in random order */
  perm = permute(helosCount);
  for(i=0;i<helosCount;i++) {
    addCronJob((CronJob)&receiveHeloDeferred,		 
	        15 * cronSECONDS * (i+1),
		0,
		helos[perm[i]]);
  }  
  FREE(perm);
  GROW(helos,
       helosCount,
       0);
  FREE(buffer);
  CLOSE(sock);  
#if DEBUG_HELOEXCHANGE
  LOG(LOG_INFO,
      "INFO: exit downloadHostlist (%d seconds before timeout)\n",
      (int)(start + 30 * cronSECONDS - cronTime(NULL))/cronSECONDS);
#endif
}

void downloadHostlist() {
  char * url;
  int i;
  
  url = getConfigurationString("GNUNETD",
			       "HOSTLISTURL");
  if (url == NULL) {
#if DEBUG_HELOEXCHANGE
    LOG(LOG_CRON,
	"CRON: exit downloadHostlist (error: URL not specified)\n");
#endif
    return;
  }
  i = strlen(url);
  while (i > 0) {
    i--;
    if (url[i] == ' ') {
      downloadHostlistHelper(&url[i+1]);
      url[i] = '\0';
    }
  }
  downloadHostlistHelper(&url[0]);
  FREE(url);
}

typedef struct {
  /* the CRC of the message */
  int crc;
  /* the HELO message */
  HELO_Message * m;
  /* send the HELO in 1 out of n cases */
  int n;
} SendData;

static void broadcastHelper(const HostIdentity * hi,
			    const unsigned short proto,
			    SendData * sd) {
  HELO_Message * helo;
  TSession * tsession;
  HexName other;

  if (randomi(sd->n) != 0) 
    return;
  hash2hex(&hi->hashPubKey,
	   &other);
#if DEBUG_HELOEXCHANGE
  LOG(LOG_DEBUG,
      "DEBUG: enter heloexchange - broadcast helper to %s\n",
      (char*)&other);
#endif
  if (hostIdentityEquals(hi,
			 &myIdentity))
    return; /* never advertise to myself... */
  if (YES == isConnected(hi)) {
    sendToNode(hi,
	       &sd->m->header,
	       getConnectPriority(),
	       HELO_BROADCAST_FREQUENCY);
    statChange(stat_helo_initiated, 1);
    return;
  }
  /* with even lower probability: */
  if (randomi(sd->n) != 0)
    return;
	
  /* establish short-lived connection, send, tear down */
  if (SYSERR == identity2Helo(hi,
			      proto,
			      NO,
			      &helo)) {
#if DEBUG_HELOEXCHANGE
    LOG(LOG_DEBUG,
	"DEBUG: exit heloexchange - "
	"broadcast helper (error: identity2helo failed)\n");
#endif
    return;
  }
  if (SYSERR == transportConnect(helo, /* helo is freed by callee,
					  except on SYSERR! */
				 &tsession)) {
    FREE(helo);
#if DEBUG_HELOEXCHANGE
    LOG(LOG_DEBUG,
	"DEBUG: enter heloexchange - "\
	"broadcast helper (connect error)\n");
#endif
    return; /* could not connect */
  }
  transportSend(tsession,
		&sd->m->header,
		HELO_Message_size(sd->m),
		NO, /* not encrypted */
		sd->crc);
  statChange(stat_helo_initiated, 1);
  transportDisconnect(tsession);
#if DEBUG_HELOEXCHANGE
  LOG(LOG_DEBUG,
      "DEBUG: exit heloexchange - broadcast helper\n");
#endif
 }

/**
 * Tell a couple of random hosts on the currentKnownHost list 
 * that we exist (called for each transport)...
 **/
static void broadcastHELOTransport(TransportAPI * tapi,
				   void * unused) {
  int j;
  SendData sd;

#if DEBUG_HELOEXCHANGE
  LOG(LOG_CRON,
      "CRON: enter broadcastHELO\n");
#endif
  j = forEachHost(NULL, 
		  0, 
		  NULL); /* just count */
  if (j < 1) {
    LOG(LOG_WARNING,
	"WARNING: announcing ourselves pointless: no hosts known\n");
    return; /* no point in trying... */
  }
  sd.n = j/((int) log(j+1)+1); /* send the HELO to log(known host) hosts */  
  if (sd.n == 0)
    sd.n = 1;
  if (SYSERR == transportCreateHELO(tapi->protocolNumber,
				    &sd.m))
    return;
#if DEBUG_HELOEXCHANGE
  LOG(LOG_INFO,
      "INFO: advertising my transport %d to selected peers\n",
      tapi->protocolNumber);
#endif
  bindAddress(sd.m);
  sd.crc = crc32N(sd.m, 
		  HELO_Message_size(sd.m));

  forEachHost((HostIterator)&broadcastHelper,
	      0, /* ignore blacklisting */
	      &sd);
  FREE(sd.m);
#if DEBUG_HELOEXCHANGE
  LOG(LOG_CRON,
      "CRON: exit broadcastHELO\n");
#endif
}

/**
 * Tell a couple of random hosts on the currentKnownHost list 
 * that we exist...
 **/
static void broadcastHELO(void * unused) {
  forEachTransport(&broadcastHELOTransport, NULL);
}


/**
 * Forward HELOs from all known hosts to all connected hosts.
 **/
static void forwardHELOHelper(const HostIdentity * identity,
    const unsigned short protocol,
    int * count) {
HELO_Message * helo;
TIME_T now;
  int ran;

  ran = log(*count+1);
  if (ran > 0)
    ran = randomi(ran);
  if (ran > randomi((*count)+1))
    return; /* only forward with a certain chance */
#if DEBUG_HELOEXCHANGE
  LOG(LOG_CRON,
      "CRON: forwarding HELOs\n");
#endif
  if (SYSERR == identity2Helo(identity,
			      protocol,
			      NO,
			      &helo))
    return; /* this should not happen */
  helo->header.requestType = htons(p2p_PROTO_HELO); 
  helo->header.size = htons(HELO_Message_size(helo));
  /* do not forward expired HELOs */
  TIME(&now);
  if ((TIME_T)ntohl(helo->expirationTime) < now) {
    HexName hex;
    /* remove HELOs that expired */ 
    IFLOG(LOG_INFO,
	  hash2hex(&identity->hashPubKey,
		   &hex));
    LOG(LOG_INFO,
	"INFO: Removing expired HELO from %s (expired %ds ago)\n",
	&hex, now - ntohl(helo->expirationTime));
    delHostFromKnown(identity, protocol);
    FREE(helo);
    return;
  }
  statChange(stat_helo_forwarded, 
	     forEachConnectedNode(NULL, NULL));
  broadcast(&helo->header, 
	    0, 
	    HELO_BROADCAST_FREQUENCY);  /* send before the next round... */
  FREE(helo);
}

/**
 * Forward HELOs from all known hosts to all connected hosts.
 **/
static void forwardHELO(void * unused) {
  int count;

#if DEBUG_HELOEXCHANGE
  LOG(LOG_CRON,
      "CRON: enter forwardHELO\n");
#endif
  count = forEachHost(NULL, 
		      0, 
		      NULL);  
  forEachHost((HostIterator)&forwardHELOHelper,
	      0, /* ignore blacklisting */
	      &count);
#if DEBUG_HELOEXCHANGE
  LOG(LOG_CRON,
      "CRON: exit forwardHELO\n");
#endif
}

/* end of heloexchange.c */
